@@ -1,16 +1,16 @@
-use crate::{Card, Result};
+use crate::{Card, Result, Uuid};
 
 pub struct App {
-    card: Card,
+    pub card: Card,
 }
 
 impl super::App for App {
     const RID: &'static [u8] = super::NIST_RID;
     const PIX: &'static [u8] = super::PIV_PIX;
 
-    fn new() -> Result<Self> {
+    fn new(uuid: Option<Uuid>) -> Result<Self> {
         Ok(Self {
-            card: Self::connect()?,
+            card: Self::connect(uuid)?,
         })
     }
 
